'use strict';

var Options = require('../../../../Options.js');
var getCleanPasswords = require('./unmunger/getCleanPasswords.js');

const getExtras = (passwordWithSubs, token) => {
  const usedChanges = passwordWithSubs.changes.filter(changes => {
    const sub = changes.substitution;
    return token.indexOf(sub) !== -1;
  });
  const subDisplay = [];
  const filtered = usedChanges.filter((value, index, self) => {
    const existingIndex = self.findIndex(t => {
      return t.letter === value.letter && t.substitution === value.substitution;
    });
    return index === existingIndex;
  });
  filtered.forEach(value => {
    subDisplay.push(`${value.substitution} -> ${value.letter}`);
  });
  return {
    subs: filtered,
    subDisplay: subDisplay.join(', ')
  };
};
/*
 * -------------------------------------------------------------------------------
 *  Dictionary l33t matching -----------------------------------------------------
 * -------------------------------------------------------------------------------
 */
class MatchL33t {
  constructor(defaultMatch) {
    this.defaultMatch = defaultMatch;
  }
  isAlreadyIncluded(matches, newMatch) {
    return matches.some(l33tMatch => {
      return Object.entries(l33tMatch).every(([key, value]) => {
        return key === 'subs' || value === newMatch[key];
      });
    });
  }
  match({
    password
  }) {
    const matches = [];
    const subbedPasswords = getCleanPasswords(password, Options.zxcvbnOptions.l33tMaxSubstitutions, Options.zxcvbnOptions.trieNodeRoot);
    let hasFullMatch = false;
    let isFullSubstitution = true;
    subbedPasswords.reverse().forEach(subbedPassword => {
      if (hasFullMatch) {
        return;
      }
      const matchedDictionary = this.defaultMatch({
        password: subbedPassword.password,
        useLevenshtein: isFullSubstitution
      });
      // only the first entry has a full substitution
      isFullSubstitution = false;
      matchedDictionary.forEach(match => {
        if (!hasFullMatch) {
          hasFullMatch = match.i === 0 && match.j === password.length - 1;
        }
        const token = password.slice(match.i, +match.j + 1 || 9e9);
        const extras = getExtras(subbedPassword, token);
        const newMatch = {
          ...match,
          l33t: true,
          token,
          ...extras
        };
        const alreadyIncluded = this.isAlreadyIncluded(matches, newMatch);
        // only return the matches that contain an actual substitution
        if (token.toLowerCase() !== match.matchedWord && !alreadyIncluded) {
          matches.push(newMatch);
        }
      });
    });
    // filter single-character l33t matches to reduce noise.
    // otherwise '1' matches 'i', '4' matches 'a', both very common English words
    // with low dictionary rank.
    return matches.filter(match => match.token.length > 1);
  }
}

module.exports = MatchL33t;
//# sourceMappingURL=l33t.js.map
